package configversion

import (
	"context"
	"fmt"

	"github.com/tofutf/tofutf/internal/rbac"
)

// cacheKey generates a key for caching config tarballs
func cacheKey(cvID string) string {
	return fmt.Sprintf("%s.tar.gz", cvID)
}

// UploadConfig saves a configuration tarball to the db
//
// NOTE: unauthenticated - access granted only via signed URL
func (s *Service) UploadConfig(ctx context.Context, cvID string, config []byte) error {
	err := s.db.UploadConfigurationVersion(ctx, cvID, func(cv *ConfigurationVersion, uploader ConfigUploader) error {
		return cv.Upload(ctx, config, uploader)
	})
	if err != nil {
		s.logger.Error("uploading configuration", "err", err)
		return err
	}

	if err := s.cache.Set(cacheKey(cvID), config); err != nil {
		s.logger.Error("caching configuration version tarball", "err", err)
	}

	s.logger.Info("uploaded configuration", "id", cvID, "bytes", len(config))
	return nil
}

// DownloadConfig retrieves a tarball from the db
func (s *Service) DownloadConfig(ctx context.Context, cvID string) ([]byte, error) {
	subject, err := s.canAccess(ctx, rbac.DownloadConfigurationVersionAction, cvID)
	if err != nil {
		return nil, err
	}

	if config, err := s.cache.Get(cacheKey(cvID)); err == nil {
		return config, nil
	}
	config, err := s.db.GetConfig(ctx, cvID)
	if err != nil {
		s.logger.Error("downloading configuration", "id", cvID, "subject", subject, "err", err)
		return nil, err
	}

	if err := s.cache.Set(cacheKey(cvID), config); err != nil {
		s.logger.Error("caching configuration version tarball", "err", err)
	}

	s.logger.Debug("downloaded configuration", "id", cvID, "bytes", len(config), "subject", subject)

	return config, nil
}
