package integration

import (
	"bytes"
	"context"
	"io"
	"log/slog"
	"net/url"
	"os"
	"os/exec"
	"testing"

	"github.com/google/uuid"
	"github.com/stretchr/testify/require"
	"github.com/tofutf/tofutf/internal"
	"github.com/tofutf/tofutf/internal/agent"
	"github.com/tofutf/tofutf/internal/api"
	"github.com/tofutf/tofutf/internal/cli"
	"github.com/tofutf/tofutf/internal/configversion"
	"github.com/tofutf/tofutf/internal/daemon"
	"github.com/tofutf/tofutf/internal/github"
	"github.com/tofutf/tofutf/internal/module"
	"github.com/tofutf/tofutf/internal/notifications"
	"github.com/tofutf/tofutf/internal/organization"
	"github.com/tofutf/tofutf/internal/releases"
	"github.com/tofutf/tofutf/internal/run"
	"github.com/tofutf/tofutf/internal/sql"
	"github.com/tofutf/tofutf/internal/state"
	"github.com/tofutf/tofutf/internal/team"
	otfuser "github.com/tofutf/tofutf/internal/user"
	"github.com/tofutf/tofutf/internal/variable"
	"github.com/tofutf/tofutf/internal/vcs"
	"github.com/tofutf/tofutf/internal/vcsprovider"
	"github.com/tofutf/tofutf/internal/workspace"
	xslog "github.com/tofutf/tofutf/internal/xslog"
)

type (
	// daemon for integration test
	testDaemon struct {
		*daemon.Daemon
		// stub github server for test to use.
		*github.TestServer
		// dowloader allows tests to download terraform
		downloader
	}

	// downloader downloads terraform versions
	downloader interface {
		Download(ctx context.Context, version string, w io.Writer) (string, error)
	}

	// configures the daemon for integration tests
	config struct {
		daemon.Config
		// skip creation of default organization
		skipDefaultOrganization bool
		// customise path in which terraform bins are saved
		terraformBinDir string
	}
)

// setup an integration test with a daemon, organization, and a user context.
func setup(t *testing.T, cfg *config, gopts ...github.TestServerOption) (*testDaemon, *organization.Organization, context.Context) {
	t.Helper()

	if cfg == nil {
		cfg = &config{}
	}
	// Setup database if not specified
	if cfg.Database == "" {
		cfg.Database = sql.NewTestDB(t)
	}
	// Use shared secret if one is not specified
	if cfg.Secret == nil {
		cfg.Secret = sharedSecret
	}
	// Unless test has specified otherwise, disable checking for latest
	// terraform version
	if cfg.DisableLatestChecker == nil || !*cfg.DisableLatestChecker {
		cfg.DisableLatestChecker = internal.Bool(true)
	}
	// Skip TLS verification for tests because they'll be standing up various
	// stub TLS servers with self-certified certs.
	cfg.SkipTLSVerification = true

	daemon.ApplyDefaults(&cfg.Config)
	cfg.SSL = true
	cfg.CertFile = "./fixtures/cert.pem"
	cfg.KeyFile = "./fixtures/key.pem"

	// Start stub github server, unless test has set its own github stub
	var githubServer *github.TestServer
	if cfg.GithubHostname == "" {
		var githubURL *url.URL
		githubServer, githubURL = github.NewTestServer(t, gopts...)
		cfg.GithubHostname = githubURL.Host
	}

	// Configure logger; discard logs by default
	var logger *slog.Logger
	if _, ok := os.LookupEnv("OTF_INTEGRATION_TEST_ENABLE_LOGGER"); ok {
		var err error
		logger = slog.New(slog.NewTextHandler(os.Stderr, &slog.HandlerOptions{Level: slog.LevelDebug}))
		require.NoError(t, err)
	} else {
		logger = slog.New(&xslog.NoopHandler{})
	}

	// Confer superuser privileges on all calls to service endpoints
	ctx := internal.AddSubjectToContext(context.Background(), &internal.Superuser{Username: "app-user"})

	d, err := daemon.New(ctx, logger, cfg.Config)
	require.NoError(t, err)

	// start daemon and upon test completion check that it exited cleanly
	ctx, cancel := context.WithCancel(ctx)
	done := make(chan error)
	started := make(chan struct{})
	go func() {
		err := d.Start(ctx, started)
		// if context was canceled don't report any error
		if ctx.Err() != nil {
			done <- nil
			return
		}
		require.NoError(t, err, "daemon exited with an error")
		done <- err
	}()
	// don't proceed until daemon has started.
	<-started

	t.Cleanup(func() {
		cancel() // terminates daemon
		<-done   // don't exit test until daemon is fully terminated
	})

	daemon := &testDaemon{
		Daemon:     d,
		TestServer: githubServer,
		downloader: releases.NewDownloader(cfg.terraformBinDir),
	}

	// create a dedicated user account and context for test to use.
	testUser, testUserCtx := daemon.createUserCtx(t)

	var org *organization.Organization
	if !cfg.skipDefaultOrganization {
		// create organization for test to use. Consume the created event too so
		// that tests that consume events don't receive this event.
		org = daemon.createOrganization(t, testUserCtx)
		// re-fetch user so that its ownership of the above org is included
		testUser = daemon.getUser(t, adminCtx, testUser.Username)
		// and re-add to context
		testUserCtx = internal.AddSubjectToContext(ctx, testUser)
	}

	return daemon, org, testUserCtx
}

func (s *testDaemon) createOrganization(t *testing.T, ctx context.Context) *organization.Organization {
	t.Helper()

	org, err := s.Organizations.Create(ctx, organization.CreateOptions{
		Name: internal.String(internal.GenerateRandomString(4) + "-corp"),
	})
	require.NoError(t, err)
	return org
}

func (s *testDaemon) createWorkspace(t *testing.T, ctx context.Context, org *organization.Organization) *workspace.Workspace {
	t.Helper()

	if org == nil {
		org = s.createOrganization(t, ctx)
	}

	ws, err := s.Workspaces.Create(ctx, workspace.CreateOptions{
		Name:         internal.String("workspace-" + internal.GenerateRandomString(6)),
		Organization: &org.Name,
	})
	require.NoError(t, err)
	return ws
}

func (s *testDaemon) getWorkspace(t *testing.T, ctx context.Context, workspaceID string) *workspace.Workspace {
	t.Helper()

	ws, err := s.Workspaces.Get(ctx, workspaceID)
	require.NoError(t, err)
	return ws
}

func (s *testDaemon) createVCSProvider(t *testing.T, ctx context.Context, org *organization.Organization) *vcsprovider.VCSProvider {
	t.Helper()

	if org == nil {
		org = s.createOrganization(t, ctx)
	}

	provider, err := s.VCSProviders.Create(ctx, vcsprovider.CreateOptions{
		Organization: org.Name,
		// tests require a legitimate cloud name to avoid invalid foreign
		// key error upon insert/update
		Kind:  vcs.KindPtr(vcs.GithubKind),
		Token: internal.String(uuid.NewString()),
	})
	require.NoError(t, err)
	return provider
}

func (s *testDaemon) createModule(t *testing.T, ctx context.Context, org *organization.Organization) *module.Module {
	t.Helper()

	if org == nil {
		org = s.createOrganization(t, ctx)
	}

	module, err := s.Modules.CreateModule(ctx, module.CreateOptions{
		Name:         uuid.NewString(),
		Provider:     uuid.NewString(),
		Organization: org.Name,
	})
	require.NoError(t, err)
	return module
}

// createUser is always invoked with the site admin context because only they
// are authorized to create users.
func (s *testDaemon) createUser(t *testing.T, opts ...otfuser.NewUserOption) *otfuser.User {
	t.Helper()

	user, err := s.Users.Create(adminCtx, "user-"+internal.GenerateRandomString(4), opts...)
	require.NoError(t, err)
	return user
}

// createUserCtx is always invoked with the site admin context because only they
// are authorized to create users.
func (s *testDaemon) createUserCtx(t *testing.T, opts ...otfuser.NewUserOption) (*otfuser.User, context.Context) {
	t.Helper()

	user := s.createUser(t, opts...)
	return user, internal.AddSubjectToContext(context.Background(), user)
}

func (s *testDaemon) getUser(t *testing.T, ctx context.Context, username string) *otfuser.User {
	t.Helper()

	user, err := s.Users.GetUser(ctx, otfuser.UserSpec{Username: &username})
	require.NoError(t, err)
	return user
}

func (s *testDaemon) getUserCtx(t *testing.T, ctx context.Context, username string) (*otfuser.User, context.Context) {
	t.Helper()

	user, err := s.Users.GetUser(ctx, otfuser.UserSpec{Username: &username})
	require.NoError(t, err)
	return user, internal.AddSubjectToContext(ctx, user)
}

func (s *testDaemon) createTeam(t *testing.T, ctx context.Context, org *organization.Organization) *team.Team {
	t.Helper()

	if org == nil {
		org = s.createOrganization(t, ctx)
	}

	team, err := s.Teams.Create(ctx, org.Name, team.CreateTeamOptions{
		Name: internal.String("team-" + internal.GenerateRandomString(4)),
	})
	require.NoError(t, err)
	return team
}

func (s *testDaemon) getTeam(t *testing.T, ctx context.Context, org, name string) *team.Team {
	t.Helper()

	team, err := s.Teams.Get(ctx, org, name)
	require.NoError(t, err)
	return team
}

func (s *testDaemon) createConfigurationVersion(t *testing.T, ctx context.Context, ws *workspace.Workspace, opts *configversion.CreateOptions) *configversion.ConfigurationVersion {
	t.Helper()

	if ws == nil {
		ws = s.createWorkspace(t, ctx, nil)
	}
	if opts == nil {
		opts = &configversion.CreateOptions{}
	}

	cv, err := s.Configs.Create(ctx, ws.ID, *opts)
	require.NoError(t, err)
	return cv
}

func (s *testDaemon) createAndUploadConfigurationVersion(t *testing.T, ctx context.Context, ws *workspace.Workspace, opts *configversion.CreateOptions) *configversion.ConfigurationVersion {
	cv := s.createConfigurationVersion(t, ctx, ws, opts)
	tarball, err := os.ReadFile("./testdata/root.tar.gz")
	require.NoError(t, err)
	err = s.Configs.UploadConfig(ctx, cv.ID, tarball)
	require.NoError(t, err)
	return cv
}

func (s *testDaemon) createRun(t *testing.T, ctx context.Context, ws *workspace.Workspace, cv *configversion.ConfigurationVersion) *run.Run {
	t.Helper()

	if ws == nil {
		ws = s.createWorkspace(t, ctx, nil)
	}
	if cv == nil {
		cv = s.createConfigurationVersion(t, ctx, ws, nil)
	}

	run, err := s.Runs.Create(ctx, ws.ID, run.CreateOptions{
		ConfigurationVersionID: internal.String(cv.ID),
	})
	require.NoError(t, err)
	return run
}

func (s *testDaemon) createVariable(t *testing.T, ctx context.Context, ws *workspace.Workspace) *variable.Variable {
	t.Helper()

	if ws == nil {
		ws = s.createWorkspace(t, ctx, nil)
	}

	v, err := s.Variables.CreateWorkspaceVariable(ctx, ws.ID, variable.CreateVariableOptions{
		Key:      internal.String("key-" + internal.GenerateRandomString(4)),
		Value:    internal.String("val-" + internal.GenerateRandomString(4)),
		Category: variable.VariableCategoryPtr(variable.CategoryTerraform),
	})
	require.NoError(t, err)
	return v
}

func (s *testDaemon) createStateVersion(t *testing.T, ctx context.Context, ws *workspace.Workspace) *state.Version {
	t.Helper()

	if ws == nil {
		ws = s.createWorkspace(t, ctx, nil)
	}

	file, err := os.ReadFile("./testdata/terraform.tfstate")
	require.NoError(t, err)

	sv, err := s.State.Create(ctx, state.CreateStateVersionOptions{
		State:       file,
		WorkspaceID: internal.String(ws.ID),
		// serial matches that in ./testdata/terraform.tfstate
		Serial: internal.Int64(9),
	})
	require.NoError(t, err)
	return sv
}

func (s *testDaemon) getCurrentState(t *testing.T, ctx context.Context, wsID string) *state.Version {
	t.Helper()

	sv, err := s.State.GetCurrent(ctx, wsID)
	require.NoError(t, err)
	return sv
}

func (s *testDaemon) createToken(t *testing.T, ctx context.Context, user *otfuser.User) (*otfuser.UserToken, []byte) {
	t.Helper()

	// If user is provided then add them to context. Otherwise the context is
	// expected to contain a user if authz is to succeed.
	if user != nil {
		ctx = internal.AddSubjectToContext(ctx, user)
	}

	ut, token, err := s.Users.CreateToken(ctx, otfuser.CreateUserTokenOptions{
		Description: "lorem ipsum...",
	})
	require.NoError(t, err)
	return ut, token
}

func (s *testDaemon) createNotificationConfig(t *testing.T, ctx context.Context, ws *workspace.Workspace) *notifications.Config {
	t.Helper()

	if ws == nil {
		ws = s.createWorkspace(t, ctx, nil)
	}

	nc, err := s.Notifications.Create(ctx, ws.ID, notifications.CreateConfigOptions{
		DestinationType: notifications.DestinationGeneric,
		Enabled:         internal.Bool(true),
		Name:            internal.String(uuid.NewString()),
		URL:             internal.String("http://example.com"),
	})
	require.NoError(t, err)
	return nc
}

// startAgent starts a pool agent, configuring it with the given organization
// and configuring it to connect to the daemon. The corresponding agent type is
// returned once registered, along with a function to shutdown the agent down.
func (s *testDaemon) startAgent(t *testing.T, ctx context.Context, org, poolID, token string, cfg agent.Config) (*agent.Agent, func()) {
	t.Helper()

	// Configure logger; discard logs by default
	var logger *slog.Logger
	if _, ok := os.LookupEnv("OTF_INTEGRATION_TEST_ENABLE_LOGGER"); ok {
		logger = slog.New(&xslog.NoopHandler{})
	} else {
		logger = slog.New(&xslog.NoopHandler{})
	}

	if token == "" {
		if poolID == "" {
			pool, err := s.Agents.CreateAgentPool(ctx, agent.CreateAgentPoolOptions{
				Name:         uuid.NewString(),
				Organization: org,
			})
			require.NoError(t, err)
			poolID = pool.ID
		}
		_, tokenBytes, err := s.Agents.CreateAgentToken(ctx, poolID, agent.CreateAgentTokenOptions{
			Description: "lorem ipsum...",
		})
		require.NoError(t, err)
		token = string(tokenBytes)
	}

	agentDaemon, err := agent.NewPoolDaemon(logger, cfg, api.Config{
		Token:   token,
		Address: s.System.Hostname(),
	})
	require.NoError(t, err)

	ctx, cancel := context.WithCancel(ctx)
	done := make(chan struct{})
	go func() {
		err := agentDaemon.Start(ctx)
		close(done)
		require.NoError(t, err)
	}()

	t.Cleanup(func() {
		cancel() // terminate agent
		<-done   // don't exit test until agent fully terminated
	})
	return <-agentDaemon.Registered(), cancel
}

func (s *testDaemon) tfcli(t *testing.T, ctx context.Context, command, configPath string, args ...string) string {
	t.Helper()

	out, err := s.tfcliWithError(t, ctx, command, configPath, args...)
	require.NoError(t, err, "tf cli failed: %s", out)
	return out
}

func (s *testDaemon) tfcliWithError(t *testing.T, ctx context.Context, command, configPath string, args ...string) (string, error) {
	t.Helper()

	tfpath := s.downloadTerraform(t, ctx, nil)

	// Create user token expressly for the terraform cli
	user := userFromContext(t, ctx)
	_, token := s.createToken(t, ctx, user)

	cmdargs := []string{command}
	cmdargs = append(cmdargs, args...)

	cmd := exec.Command(tfpath, cmdargs...)
	cmd.Dir = configPath

	cmd.Env = internal.SafeAppend(sharedEnvs, internal.CredentialEnv(s.System.Hostname(), token))

	// Run tf and collect stdout/stderr
	out, err := cmd.CombinedOutput()

	// strip ANSI escape codes because tests are not expecting them.
	return internal.StripAnsi(string(out)), err
}

func (s *testDaemon) otfcli(t *testing.T, ctx context.Context, args ...string) string {
	t.Helper()

	// Create user token expressly for the otf cli
	user := userFromContext(t, ctx)
	_, token := s.createToken(t, ctx, user)

	cmdargs := []string{"--address", s.System.Hostname(), "--token", string(token)}
	cmdargs = append(cmdargs, args...)

	var buf bytes.Buffer
	err := cli.NewCLI().Run(ctx, cmdargs, &buf)
	require.NoError(t, err)

	require.NoError(t, err, "otf cli failed: %s", buf.String())
	return buf.String()
}

func (s *testDaemon) downloadTerraform(t *testing.T, ctx context.Context, version *string) string {
	t.Helper()

	if version == nil {
		version = internal.String(releases.DefaultTerraformVersion)
	}
	tfpath, err := s.Download(ctx, *version, io.Discard)
	require.NoError(t, err)
	return tfpath
}
