package notifications

import (
	"context"
	"fmt"
	"log/slog"

	"github.com/tofutf/tofutf/internal"
	"github.com/tofutf/tofutf/internal/pubsub"
	"github.com/tofutf/tofutf/internal/run"
	"github.com/tofutf/tofutf/internal/sql"
	"github.com/tofutf/tofutf/internal/workspace"
)

// LockID guarantees only one notifier on a cluster is running at any
// time.
const LockID int64 = 5577006791947779411

type (
	// Notifier relays run events onto interested parties
	Notifier struct {
		logger *slog.Logger

		workspaces    notifierWorkspaceClient
		runs          notifierRunClient
		notifications notifierNotificationClient
		system        notifierHostnameClient

		*cache
		db *pgdb
	}

	NotifierOptions struct {
		RunClient          notifierRunClient
		WorkspaceClient    notifierWorkspaceClient
		NotificationClient notifierNotificationClient
		Logger             *slog.Logger

		*internal.HostnameService
		*sql.Pool
	}

	notifierWorkspaceClient interface {
		Get(ctx context.Context, workspaceID string) (*workspace.Workspace, error)
	}

	notifierRunClient interface {
		Watch(context.Context) (<-chan pubsub.Event[*run.Run], func())
	}

	notifierNotificationClient interface {
		Watch(context.Context) (<-chan pubsub.Event[*Config], func())
	}

	notifierHostnameClient interface {
		Hostname() string
	}
)

func NewNotifier(opts NotifierOptions) *Notifier {
	return &Notifier{
		logger:        opts.Logger.With("component", "notifier"),
		workspaces:    opts.WorkspaceClient,
		system:        opts.HostnameService,
		runs:          opts.RunClient,
		notifications: opts.NotificationClient,
		db:            &pgdb{opts.Pool},
	}
}

// Start the notifier daemon. Should be started in a go-routine.
func (s *Notifier) Start(ctx context.Context) error {
	// subscribe to both run events and notification config events
	subRuns, unsubRuns := s.runs.Watch(ctx)
	defer unsubRuns()
	subConfigs, unsubConfigs := s.notifications.Watch(ctx)
	defer unsubConfigs()

	// populate cache with existing notification configs
	cache, err := newCache(ctx, s.db, &defaultFactory{})
	if err != nil {
		return err
	}
	s.cache = cache

	// block on handling events
	for {
		select {
		case event, ok := <-subRuns:
			if !ok {
				return pubsub.ErrSubscriptionTerminated
			}
			if err := s.handleRun(ctx, event.Payload); err != nil {
				s.logger.Error("handling event", "event", event.Type, "err", err)
			}
		case event, ok := <-subConfigs:
			if !ok {
				return pubsub.ErrSubscriptionTerminated
			}
			if err := s.handleConfig(ctx, event); err != nil {
				s.logger.Error("handling event", "event", event.Type, "err", err)
			}
		}
	}
}

func (s *Notifier) handleConfig(ctx context.Context, event pubsub.Event[*Config]) error {
	switch event.Type {
	case pubsub.CreatedEvent:
		return s.add(event.Payload)
	case pubsub.UpdatedEvent:
		if err := s.remove(event.Payload.ID); err != nil {
			return err
		}
		return s.add(event.Payload)
	case pubsub.DeletedEvent:
		return s.remove(event.Payload.ID)
	default:
		return nil
	}
}

func (s *Notifier) handleRun(ctx context.Context, r *run.Run) error {
	if r.Queued() {
		// ignore queued events
		return nil
	}

	s.mu.Lock()
	defer s.mu.Unlock()

	var ws *workspace.Workspace
	for _, cfg := range s.configs {
		if cfg.WorkspaceID != r.WorkspaceID {
			// skip configs for other workspaces
			continue
		}
		if !cfg.Enabled {
			// skip disabled config
			continue
		}
		if len(cfg.Triggers) == 0 {
			// skip config with no triggers
			continue
		}
		trigger, matches := cfg.matchTrigger(r)
		if !matches {
			// skip config with no matching trigger
			continue
		}
		// Retrieve workspace if not already retrieved. We do this in order to
		// furnish the notification below with further information.
		//
		// TODO: this is rather expensive. We should either:
		// (a) cache workspaces, either in the notifier or upstream; or
		// (b) add workspace info to run itself
		if ws == nil {
			var err error
			ws, err = s.workspaces.Get(ctx, r.WorkspaceID)
			if err != nil {
				return err
			}
		}
		client, ok := s.clients[*cfg.URL]
		if !ok {
			// should never happen
			return fmt.Errorf("client not found for url: %s", *cfg.URL)
		}
		msg := &notification{
			run:       r,
			workspace: ws,
			trigger:   trigger,
			config:    cfg,
			hostname:  s.system.Hostname(),
		}
		s.logger.Debug("publishing notification", "notification", msg)
		if err := client.Publish(ctx, msg); err != nil {
			return err
		}
	}
	return nil
}
