package organization

import (
	"context"
	"fmt"
	"log/slog"

	"github.com/gorilla/mux"
	"github.com/tofutf/tofutf/internal"
	"github.com/tofutf/tofutf/internal/http/html"
	"github.com/tofutf/tofutf/internal/pubsub"
	"github.com/tofutf/tofutf/internal/rbac"
	"github.com/tofutf/tofutf/internal/resource"
	"github.com/tofutf/tofutf/internal/sql"
	"github.com/tofutf/tofutf/internal/sql/pggen"
	"github.com/tofutf/tofutf/internal/tfeapi"
	"github.com/tofutf/tofutf/internal/tokens"
)

type (
	Service struct {
		RestrictOrganizationCreation bool

		internal.Authorizer // authorize access to org

		logger       *slog.Logger
		db           *pgdb
		site         internal.Authorizer // authorize access to site
		web          *web
		tfeapi       *tfe
		api          *api
		tokenFactory *tokenFactory
		broker       *pubsub.Broker[*Organization]

		afterCreateHooks  []func(context.Context, *Organization) error
		beforeDeleteHooks []func(context.Context, *Organization) error
	}

	Options struct {
		RestrictOrganizationCreation bool
		TokensService                *tokens.Service

		*sql.Pool
		*tfeapi.Responder
		*sql.Listener
		html.Renderer
		Logger *slog.Logger
	}

	// ListOptions represents the options for listing organizations.
	ListOptions struct {
		resource.PageOptions
	}
)

func NewService(opts Options) *Service {
	svc := Service{
		Authorizer:                   &Authorizer{opts.Logger},
		logger:                       opts.Logger,
		RestrictOrganizationCreation: opts.RestrictOrganizationCreation,
		db:                           &pgdb{opts.Pool},
		site:                         &internal.SiteAuthorizer{Logger: opts.Logger},
		tokenFactory:                 &tokenFactory{tokens: opts.TokensService},
	}
	svc.web = &web{
		Renderer:         opts.Renderer,
		RestrictCreation: opts.RestrictOrganizationCreation,
		svc:              &svc,
	}
	svc.tfeapi = &tfe{
		Service:   &svc,
		Responder: opts.Responder,
	}

	svc.api = &api{
		Service:   &svc,
		Responder: opts.Responder,
	}
	svc.broker = pubsub.NewBroker(
		opts.Logger,
		opts.Listener,
		"organizations",
		func(ctx context.Context, id string, action sql.Action) (*Organization, error) {
			if action == sql.DeleteAction {
				return &Organization{ID: id}, nil
			}
			return svc.db.getByID(ctx, id)
		},
	)
	// Fetch organization when API calls request organization be included in the
	// response
	opts.Responder.Register(tfeapi.IncludeOrganization, svc.tfeapi.include)
	// Register with auth middleware the organization token and a means of
	// retrieving organization corresponding to token.
	opts.TokensService.RegisterKind(OrganizationTokenKind, func(ctx context.Context, tokenID string) (internal.Subject, error) {
		return svc.getOrganizationTokenByID(ctx, tokenID)
	})
	return &svc
}

func (s *Service) AddHandlers(r *mux.Router) {
	s.web.addHandlers(r)
	s.tfeapi.addHandlers(r)
	s.api.addHandlers(r)
}

func (s *Service) WatchOrganizations(ctx context.Context) (<-chan pubsub.Event[*Organization], func()) {
	return s.broker.Subscribe(ctx)
}

// Create creates an organization. Only users can create
// organizations, or, if RestrictOrganizationCreation is true, then only the
// site admin can create organizations. Creating an organization automatically
// creates an owners team and adds creator as an owner.
func (s *Service) Create(ctx context.Context, opts CreateOptions) (*Organization, error) {
	creator, err := s.restrictOrganizationCreation(ctx)
	if err != nil {
		return nil, err
	}

	org, err := NewOrganization(opts)
	if err != nil {
		return nil, fmt.Errorf("creating organization: %w", err)
	}

	err = s.db.Tx(ctx, func(ctx context.Context, q pggen.Querier) error {
		if err := s.db.create(ctx, org); err != nil {
			return err
		}
		for _, hook := range s.afterCreateHooks {
			if err := hook(ctx, org); err != nil {
				return err
			}
		}
		return nil
	})
	if err != nil {
		s.logger.Error("creating organization", "id", org.ID, "subject", creator, "err", err)
		return nil, sql.Error(err)
	}
	s.logger.Info("created organization", "id", org.ID, "name", org.Name, "subject", creator)

	return org, nil
}

func (s *Service) AfterCreateOrganization(hook func(context.Context, *Organization) error) {
	s.afterCreateHooks = append(s.afterCreateHooks, hook)
}

func (s *Service) Update(ctx context.Context, name string, opts UpdateOptions) (*Organization, error) {
	subject, err := s.CanAccess(ctx, rbac.UpdateOrganizationAction, name)
	if err != nil {
		return nil, err
	}

	org, err := s.db.update(ctx, name, func(org *Organization) error {
		return org.Update(opts)
	})
	if err != nil {
		s.logger.Error("updating organization", "name", name, "subject", subject, "err", err)
		return nil, err
	}

	s.logger.Info("updated organization", "name", name, "id", org.ID, "subject", subject)

	return org, nil
}

// List lists organizations according to the subject. If the
// subject has site-wide permission to list organizations then all organizations
// are listed. Otherwise:
// Subject is a user: list their organization memberships
// Subject is an agent: return its organization
// Subject is an organization token: return its organization
// Subject is a team: return its organization
func (s *Service) List(ctx context.Context, opts ListOptions) (*resource.Page[*Organization], error) {
	subject, err := internal.SubjectFromContext(ctx)
	if err != nil {
		return nil, err
	}
	if subject.CanAccessSite(rbac.ListOrganizationsAction) {
		return s.db.list(ctx, dbListOptions{PageOptions: opts.PageOptions})
	}
	return s.db.list(ctx, dbListOptions{PageOptions: opts.PageOptions, names: subject.Organizations()})
}

func (s *Service) Get(ctx context.Context, name string) (*Organization, error) {
	subject, err := s.CanAccess(ctx, rbac.GetOrganizationAction, name)
	if err != nil {
		return nil, err
	}

	org, err := s.db.get(ctx, name)
	if err != nil {
		s.logger.Error("retrieving organization", "name", name, "subject", subject, "err", err)
		return nil, err
	}

	s.logger.Debug("retrieved organization", "name", name, "subject", subject)

	return org, nil
}

func (s *Service) Delete(ctx context.Context, name string) error {
	subject, err := s.CanAccess(ctx, rbac.DeleteOrganizationAction, name)
	if err != nil {
		return err
	}

	err = s.db.Tx(ctx, func(ctx context.Context, q pggen.Querier) error {
		org, err := s.db.get(ctx, name)
		if err != nil {
			return err
		}
		for _, hook := range s.beforeDeleteHooks {
			if err := hook(ctx, org); err != nil {
				return err
			}
		}
		return s.db.delete(ctx, name)
	})
	if err != nil {
		s.logger.Error("deleting organization", "name", name, "subject", subject, "err", err)
		return err
	}
	s.logger.Info("deleted organization", "name", name, "subject", subject)

	return nil
}

func (s *Service) BeforeDeleteOrganization(hook func(context.Context, *Organization) error) {
	s.beforeDeleteHooks = append(s.beforeDeleteHooks, hook)
}

func (s *Service) GetEntitlements(ctx context.Context, organization string) (Entitlements, error) {
	_, err := s.CanAccess(ctx, rbac.GetEntitlementsAction, organization)
	if err != nil {
		return Entitlements{}, err
	}

	org, err := s.Get(ctx, organization)
	if err != nil {
		return Entitlements{}, err
	}
	return defaultEntitlements(org.ID), nil
}

func (s *Service) restrictOrganizationCreation(ctx context.Context) (internal.Subject, error) {
	subject, err := internal.SubjectFromContext(ctx)
	if err != nil {
		return nil, err
	}
	if s.RestrictOrganizationCreation && !subject.IsSiteAdmin() {
		s.logger.Error("unauthorized action", "action", rbac.CreateOrganizationAction, "subject", subject)
		return subject, internal.ErrAccessNotPermitted
	}
	return subject, nil
}

// CreateToken creates an organization token. If an organization
// token already exists it is replaced.
func (s *Service) CreateToken(ctx context.Context, opts CreateOrganizationTokenOptions) (*OrganizationToken, []byte, error) {
	_, err := s.CanAccess(ctx, rbac.CreateOrganizationTokenAction, opts.Organization)
	if err != nil {
		return nil, nil, err
	}

	ot, token, err := s.tokenFactory.NewOrganizationToken(opts)
	if err != nil {
		s.logger.Error("constructing organization token", "organization", opts.Organization, "err", err)
		return nil, nil, err
	}

	if err := s.db.upsertOrganizationToken(ctx, ot); err != nil {
		s.logger.Error("creating organization token", "organization", opts.Organization, "err", err)
		return nil, nil, err
	}

	s.logger.Info("created organization token", "organization", opts.Organization)

	return ot, token, nil
}

func (s *Service) GetOrganizationToken(ctx context.Context, organization string) (*OrganizationToken, error) {
	ot, err := s.db.getOrganizationTokenByName(ctx, organization)
	if err != nil {
		s.logger.Error("retrieving organization token", "organization", organization, "err", err)
		return nil, err
	}
	s.logger.Info("retrieved organization token", "organization", organization)
	return ot, nil
}

func (s *Service) getOrganizationTokenByID(ctx context.Context, tokenID string) (*OrganizationToken, error) {
	ot, err := s.db.getOrganizationTokenByID(ctx, tokenID)
	if err != nil {
		s.logger.Error("retrieving organization token", "token_id", tokenID, "err", err)
		return nil, err
	}
	s.logger.Info("retrieved organization token", "token_id", tokenID, "organization", ot.Organization)
	return ot, nil
}

func (s *Service) ListTokens(ctx context.Context, organization string) ([]*OrganizationToken, error) {
	tokens, err := s.db.listOrganizationTokens(ctx, organization)
	if err != nil {
		s.logger.Error("listing organization tokens", "organization", organization, "err", err)
		return nil, err
	}
	s.logger.Info("listed organization tokens", "organization", organization, "count", len(tokens))
	return tokens, nil
}

func (s *Service) DeleteToken(ctx context.Context, organization string) error {
	_, err := s.CanAccess(ctx, rbac.CreateOrganizationTokenAction, organization)
	if err != nil {
		return err
	}

	if err := s.db.deleteOrganizationToken(ctx, organization); err != nil {
		s.logger.Error("deleting organization token", "organization", organization, "err", err)
		return err
	}

	s.logger.Info("deleted organization token", "organization", organization)

	return nil
}
