package team

import (
	"context"
	"time"

	"log/slog"

	"github.com/tofutf/tofutf/internal"
	"github.com/tofutf/tofutf/internal/rbac"
	"github.com/tofutf/tofutf/internal/tokens"
)

const TeamTokenKind tokens.Kind = "team_token"

type (
	// Token provides information about an API token for a team.
	Token struct {
		ID        string
		CreatedAt time.Time

		// Token belongs to a team
		TeamID string
		// Optional expiry.
		Expiry *time.Time
	}

	// CreateTokenOptions are options for creating an team token via the service
	// endpoint
	CreateTokenOptions struct {
		TeamID string
		Expiry *time.Time
	}

	teamTokenFactory struct {
		tokens *tokens.Service
	}
)

func (f *teamTokenFactory) NewTeamToken(opts CreateTokenOptions) (*Token, []byte, error) {
	tt := Token{
		ID:        internal.NewID("tt"),
		CreatedAt: internal.CurrentTimestamp(nil),
		TeamID:    opts.TeamID,
		Expiry:    opts.Expiry,
	}
	token, err := f.tokens.NewToken(tokens.NewTokenOptions{
		Subject: tt.ID,
		Kind:    TeamTokenKind,
		Expiry:  opts.Expiry,
	})
	if err != nil {
		return nil, nil, err
	}
	return &tt, token, nil
}

func (t *Token) LogValue() slog.Value {
	attrs := []slog.Attr{
		slog.String("id", t.ID),
		slog.String("team_id", t.TeamID),
	}
	if t.Expiry != nil {
		attrs = append(attrs, slog.Time("expiry", *t.Expiry))
	}
	return slog.GroupValue(attrs...)
}

func (a *Service) CreateTeamToken(ctx context.Context, opts CreateTokenOptions) (*Token, []byte, error) {
	_, err := a.team.CanAccess(ctx, rbac.CreateTeamTokenAction, opts.TeamID)
	if err != nil {
		return nil, nil, err
	}

	tt, token, err := a.NewTeamToken(opts)
	if err != nil {
		a.logger.Error("constructing team token", "team_id", opts.TeamID, "err", err)
		return nil, nil, err
	}

	if err := a.db.createTeamToken(ctx, tt); err != nil {
		a.logger.Error("creating team token", "token", tt, "err", err)
		return nil, nil, err
	}

	a.logger.Info("created team token", "token", tt)

	return tt, token, nil
}

func (a *Service) GetTeamToken(ctx context.Context, teamID string) (*Token, error) {
	_, err := a.team.CanAccess(ctx, rbac.GetTeamTokenAction, teamID)
	if err != nil {
		return nil, err
	}
	return a.db.getTeamTokenByTeamID(ctx, teamID)
}

func (a *Service) DeleteTeamToken(ctx context.Context, teamID string) error {
	_, err := a.team.CanAccess(ctx, rbac.DeleteTeamTokenAction, teamID)
	if err != nil {
		return err
	}

	if err := a.db.deleteTeamToken(ctx, teamID); err != nil {
		a.logger.Error("deleting team token", "team", teamID, "err", err)
		return err
	}

	a.logger.Info("deleted team token", "team", teamID)

	return nil
}
