package workspace

import (
	"context"
	"errors"
	"fmt"

	"github.com/tofutf/tofutf/internal"
	"github.com/tofutf/tofutf/internal/rbac"
	"github.com/tofutf/tofutf/internal/resource"
	"github.com/tofutf/tofutf/internal/sql/pggen"
)

type (
	// ListTagsOptions are options for paginating and filtering a list of
	// tags
	ListTagsOptions struct {
		resource.PageOptions
	}

	// ListWorkspaceTagsOptions are options for paginating and filtering a list of
	// workspace tags
	ListWorkspaceTagsOptions struct {
		resource.PageOptions
	}
)

func (s *Service) ListTags(ctx context.Context, organization string, opts ListTagsOptions) (*resource.Page[*Tag], error) {
	subject, err := s.organization.CanAccess(ctx, rbac.ListTagsAction, organization)
	if err != nil {
		return nil, err
	}

	list, err := s.db.listTags(ctx, organization, opts)
	if err != nil {
		s.logger.Error("listing tags", "organization", organization, "subject", subject, "err", err)
	}
	s.logger.Debug("listed tags", "organization", organization, "subject", subject)
	return list, nil
}

func (s *Service) DeleteTags(ctx context.Context, organization string, tagIDs []string) error {
	subject, err := s.organization.CanAccess(ctx, rbac.DeleteTagsAction, organization)
	if err != nil {
		return err
	}

	if err := s.db.deleteTags(ctx, organization, tagIDs); err != nil {
		s.logger.Error("deleting tags", "organization", organization, "tags_ids", tagIDs, "subject", subject, "err", err)
		return err
	}

	s.logger.Info("deleted tags", "organization", organization, "tag_ids", tagIDs, "subject", subject)
	return nil
}

func (s *Service) TagWorkspaces(ctx context.Context, tagID string, workspaceIDs []string) error {
	subject, err := internal.SubjectFromContext(ctx)
	if err != nil {
		return err
	}

	err = s.db.Tx(ctx, func(ctx context.Context, _ pggen.Querier) error {
		for _, wid := range workspaceIDs {
			_, err := s.CanAccess(ctx, rbac.TagWorkspacesAction, wid)
			if err != nil {
				return err
			}
			if err := s.db.tagWorkspace(ctx, wid, tagID); err != nil {
				return err
			}
		}
		return nil
	})
	if err != nil {
		s.logger.Error("tagging tags", "tag_id", tagID, "workspace_ids", workspaceIDs, "subject", subject, "err", err)
		return err
	}

	s.logger.Info("tagged workspaces", "tag_id", tagID, "workspaces_ids", workspaceIDs, "subject", subject)
	return nil
}

func (s *Service) AddTags(ctx context.Context, workspaceID string, tags []TagSpec) error {
	subject, err := s.CanAccess(ctx, rbac.AddTagsAction, workspaceID)
	if err != nil {
		return err
	}

	ws, err := s.db.get(ctx, workspaceID)
	if err != nil {
		return fmt.Errorf("workspace not found; %s; %w", workspaceID, err)
	}

	added, err := s.addTags(ctx, ws, tags)
	if err != nil {
		s.logger.Error("adding tags", "workspace", workspaceID, "tags", TagSpecs(tags), "subject", subject, "err", err)
		return err
	}
	s.logger.Info("added tags", "workspace", workspaceID, "tags", added, "subject", subject)
	return nil
}

func (s *Service) RemoveTags(ctx context.Context, workspaceID string, tags []TagSpec) error {
	subject, err := s.CanAccess(ctx, rbac.RemoveTagsAction, workspaceID)
	if err != nil {
		return err
	}

	ws, err := s.db.get(ctx, workspaceID)
	if err != nil {
		return fmt.Errorf("workspace not found; %s; %w", workspaceID, err)
	}

	err = s.db.Lock(ctx, "tags", func(ctx context.Context, q pggen.Querier) (err error) {
		for _, t := range tags {
			if err := t.Valid(); err != nil {
				return err
			}
			var tag *Tag

			switch {
			case t.Name != "":
				tag, err = s.db.findTagByName(ctx, ws.Organization, t.Name)
				if errors.Is(err, internal.ErrResourceNotFound) {
					// ignore tags that cannot be found when specified by name
					continue
				} else if err != nil {
					return err
				}
			case t.ID != "":
				tag, err = s.db.findTagByID(ctx, ws.Organization, t.ID)
				if err != nil {
					return err
				}
			default:
				return ErrInvalidTagSpec
			}
			if err := s.db.deleteWorkspaceTag(ctx, workspaceID, tag.ID); err != nil {
				return fmt.Errorf("removing tag %s from workspace %s: %w", tag.ID, workspaceID, err)
			}
		}
		return nil
	})
	if err != nil {
		s.logger.Error("removing tags", "workspace", workspaceID, "tags", TagSpecs(tags), "subject", subject, "err", err)
		return err
	}

	s.logger.Info("removed tags", "workspace", workspaceID, "tags", TagSpecs(tags), "subject", subject)
	return nil
}

func (s *Service) ListWorkspaceTags(ctx context.Context, workspaceID string, opts ListWorkspaceTagsOptions) (*resource.Page[*Tag], error) {
	subject, err := s.CanAccess(ctx, rbac.ListWorkspaceTags, workspaceID)
	if err != nil {
		return nil, err
	}

	list, err := s.db.listWorkspaceTags(ctx, workspaceID, opts)
	if err != nil {
		s.logger.Error("listing workspace tags", "workspace", workspaceID, "subject", subject, "err", err)
		return nil, err
	}
	s.logger.Debug("listed workspace tags", "workspace", workspaceID, "subject", subject)
	return list, nil
}

func (s *Service) addTags(ctx context.Context, ws *Workspace, tags []TagSpec) ([]string, error) {
	// For each tag:
	// (i) if specified by name, create new tag if it does not exist and get its ID.
	// (ii) add tag to workspace
	var added []string
	err := s.db.Lock(ctx, "tags", func(ctx context.Context, q pggen.Querier) (err error) {
		for _, t := range tags {
			if err := t.Valid(); err != nil {
				return fmt.Errorf("invalid tag: %w", err)
			}

			id := t.ID
			name := t.Name
			switch {
			case name != "":
				existing, err := s.db.findTagByName(ctx, ws.Organization, name)
				if errors.Is(err, internal.ErrResourceNotFound) {
					id = internal.NewID("tag")
					if err := s.db.addTag(ctx, ws.Organization, name, id); err != nil {
						return fmt.Errorf("adding tag: %s %w", name, err)
					}
				} else if err != nil {
					return err
				} else {
					id = existing.ID
				}
			case id != "":
				existing, err := s.db.findTagByID(ctx, ws.Organization, t.ID)
				if err != nil {
					return err
				}
				name = existing.Name
			default:
				return ErrInvalidTagSpec
			}

			if err := s.db.tagWorkspace(ctx, ws.ID, id); err != nil {
				return err
			}
			added = append(added, name)
		}
		return nil
	})
	return added, err
}
